/*
 * $Id: cilk-nd.h,v 1.1.2.14 1997/09/05 05:16:03 fengmd Exp $
 */

/*
 * $Log: cilk-nd.h,v $
 * Revision 1.1.2.14  1997/09/05 05:16:03  fengmd
 * FREE_ID changes to 2.  Fix cilk2c for Cilk_free var problem.
 *
 * Revision 1.1.2.13  1997/09/04 23:17:45  astark
 * Fixed commutativity bug by introducing NO_INLET_ID
 *
 * Revision 1.1.2.12  1997/09/03 00:56:21  astark
 * Fixed race hashing to use filename and line num as well as var
 *
 * Revision 1.1.2.11  1997/09/02 11:34:56  fengmd
 * ND_VERBOSE defaults to 0.
 *
 * Revision 1.1.2.10  1997/08/29 07:38:38  fengmd
 * Merge malloc/free checker, race report hashing, garbage collection.
 *
 * Revision 1.1.2.9  1997/06/06 16:25:53  fengmd
 * Add the Linux port of the Nondeterminator.
 *
 * Revision 1.1.2.8  1997/03/21 22:29:14  fengmd
 * Change name from spbag to spbags.
 *
 * Revision 1.1.2.7  1997/03/13 18:48:07  fengmd
 * Declare pointer constant type to avoid gcc complaining.
 *
 * Revision 1.1.2.6  1997/03/04 00:17:21  fengmd
 * ND_VERBOSE defaults to 0.
 *
 * Revision 1.1.2.5  1997/03/04 00:06:58  fengmd
 * Fix static region to accommodate the region of global initialized variable.
 *
 * Revision 1.1.2.4  1997/03/03 21:02:11  fengmd
 * Add ND_VERBOSE to show the address range of shadow space.
 *
 * Revision 1.1.2.3  1997/02/18 19:56:32  fengmd
 * Add a command-line option for the Nondeterminator's heap size.
 *
 * Revision 1.1.2.2  1997/02/15 02:46:41  fengmd
 * Add Id and Log
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

/*
 * macros to define the address range of the static region
 */
#if CILK_SOLARIS
extern _begin_Cilk_shared;
extern _end_Cilk_shared;
#define STATIC_ARENA_START      &_begin_Cilk_shared
#define STATIC_ARENA_END        &_end_Cilk_shared

#elif CILK_SOLARIS_THREADS
extern _etext;
extern _end;
#define STATIC_ARENA_START      &_etext
#define STATIC_ARENA_END        &_end

#elif CILK_IRIX_THREADS
extern _fdata;
extern _end;
#define STATIC_ARENA_START      &_fdata
#define STATIC_ARENA_END        &_end

#elif CILK_LINUX
extern _begin_Cilk_shared;
extern _end_Cilk_shared;
#define STATIC_ARENA_START      &_begin_Cilk_shared
#define STATIC_ARENA_END        &_end_Cilk_shared
#ifndef MAP_FAILED
#define MAP_FAILED		((void *) -1)
#endif

#else
#warning The Nondeterminator has only been tested on Solaris, IRIX and LINUX
#warning Please define STATIC_ARENA_START and STATIC_ARENA_END in cilk-nd.h
#endif

/* 
 * macro to display the shadow address range
 */
#define ND_VERBOSE 0

/* 
 * macros to get shadow address of reader and writer
 */
#if (CILK_SOLARIS || CILK_SOLARIS_THREADS || CILK_LINUX)

#define STATIC_READER(x) \
  ((DisjointSetMemberT *) (((unsigned int) x) + USE_SHARED_RO(static_reader_offset)))
#define STATIC_WRITER(x) \
  ((DisjointSetMemberT *) (((unsigned int) x) + USE_SHARED_RO(static_writer_offset)))
#define HEAP_READER(x) \
  ((DisjointSetMemberT *) (((unsigned int) x) - USE_SHARED_RO(both_plus_heap_size)))
#define HEAP_WRITER(x) \
  ((DisjointSetMemberT *) (((unsigned int) x) - USE_SHARED_RO(heap_size)))

#elif CILK_IRIX_THREADS

#define SHADOW_SPACE_SHIFT 0x30000000
#define STATIC_READER(x) \
  ((DisjointSetMemberT *) (((unsigned int) x) + USE_SHARED_RO(static_reader_offset) + SHADOW_SPACE_SHIFT))
#define STATIC_WRITER(x) \
  ((DisjointSetMemberT *) (((unsigned int) x) + USE_SHARED_RO(static_writer_offset) + SHADOW_SPACE_SHIFT))
#define HEAP_READER(x) \
  ((DisjointSetMemberT *) (((unsigned int) x) - USE_SHARED_RO(both_plus_heap_size) + SHADOW_SPACE_SHIFT))
#define HEAP_WRITER(x) \
  ((DisjointSetMemberT *) (((unsigned int) x) - USE_SHARED_RO(heap_size) + SHADOW_SPACE_SHIFT))

#endif

/* 
 * macro to get shadow address of var and info
 * from shadow address of reader and writer
 */
#define SHADOW_VAR(p)       \
 ((char **) (((unsigned int) p) - USE_SHARED_RO(both_size4)))
#define SHADOW_INFO(x)      \
 ((DisjointSetMemberT *) (((unsigned int) x) - USE_SHARED_RO(both_size2)))
#define SHADOW_INFO_LINE(p) (((unsigned int) *p) >> 8)
#define SHADOW_INFO_FILE(p) (((unsigned int) *p) & 0x000000FF)
     
/* 
 * SP-bags constants
 */
#define S_SET        ((DisjointSetMemberT) 1)
#define P_SET        ((DisjointSetMemberT) 0)
#define FREE_ID      ((DisjointSetMemberT) 0)
#define NO_INLET_ID  ((DisjointSetMemberT) 2) /* must be distinct from S_SET and FREE_ID */
#define FIRST_THREAD ((DisjointSetMemberT) 3)

/* 
 * the least significant 28 bits are used for procedure id, the high-order
 * 4 bits are used for storing operator and type 
 */
#define PROC_ID_BIT     28
#define MAX_PROC_NUMBER (1UL<<PROC_ID_BIT)
#define PROC_ID(x)      (((unsigned int) x) & (MAX_PROC_NUMBER - 1))
#define OP_TYPE(x)      \
  ((((unsigned int) x) & (~(MAX_PROC_NUMBER - 1))) >> PROC_ID_BIT)
#define OP_TYPE_PROC_ID(proc_id, op_type) ((op_type << PROC_ID_BIT) + proc_id)
#define DEFAULT_ND_HEAPSIZE (16U * MEGABYTE)

/* 
 * operator takes 2 lower-order bits, and the high-order 2 bits for type 
 */
#define OPERATOR_BIT      2
#define MAX_OPERATOR      (1UL<<OPERATOR_BIT)
#define CHECK_OPERATOR(x) ((x) & (MAX_OPERATOR - 1))
#define TYPE_BIT          2
#define CHECK_TYPE(x)     (((x) & (~(MAX_OPERATOR - 1))) >> OPERATOR_BIT)

/* 
 * operator value, from 0 to 2^OPERATOR_BIT - 1
 */
#define PLUSassign 0
#define DIVassign  1
#define ANDassign  2
#define EQassign   3

/* 
 * type value, from 0 to 2^TYPE_BIT - 1
 */
#define INTtype    0
#define FLOATtype  1
#define POINTtype  2
#define OTHERtype  3

#define FREEOPTYPE 0

/* 
 * structure to store list of file name 
 */
typedef struct FileListT { 
  struct FileListT *next; 
  char *filename; 
  int index;
} FileListT;

/* 
 * indices to use for initial condition versus file name 
 */
#ifdef CILK_NDL
#define INIT_INDEX 0
#define START_FILE_INDEX 1
#endif

/* 
 * structure for race hash table 
 */
typedef struct RaceHashT {
  struct RaceHashT *next;
  char *current_name, *shadow_name;
  int current_info, shadow_info;
} RaceHashT;

#define RACE_HASH_SIZE 10000

/*
 * variables defined in nondeterminator.c, and used by Cilk rts
 */
DECLARE_SHARED_RO_VARIABLE(extern char*, heap_arena_start, );
DECLARE_SHARED_RO_VARIABLE(extern char*, heap_arena_end, );
DECLARE_SHARED_RO_VARIABLE(extern size_t, heap_size, );
DECLARE_SHARED_RO_VARIABLE(extern size_t, both_plus_heap_size, );
DECLARE_SHARED_RO_VARIABLE(extern int, lreturn_size, );
DECLARE_SHARED_RO_VARIABLE(extern int, first_stop, );
DECLARE_SHARED_RO_VARIABLE(extern int, float_comm, );
DECLARE_SHARED_RO_VARIABLE(extern int, nd_arena, );
DECLARE_SHARED_RO_VARIABLE(extern DisjointSetMemberT, current_proc_id, );
#ifdef CILK_NDL
DECLARE_SHARED_RO_VARIABLE(extern int, race_linenum, );
DECLARE_SHARED_RO_VARIABLE(extern char*, race_filename, );
DECLARE_SHARED_RO_VARIABLE(extern char*, race_name, );
#endif

/* 
 * SP-bags action functions, called from cilk2c'ed Cilk program or rts
 */
extern void Cilk_init_spbags(StackFrame *f);
extern void Cilk_lreturn_spbags(StackFrame *f,volatile StackFrame *child);
extern void Cilk_sync_spbags(StackFrame *f);
extern void Cilk_alloc_proc_id_spbags();
extern inline void update_ref_count(DisjointSetMemberT *p, 
			     DisjointSetMemberT old_proc_id,
			     DisjointSetMemberT new_proc_id,
			     DisjointSetMemberT new_proc_id_type);
extern void nondeterminator_init();
#ifdef CILK_NDL
extern inline void save_shadow_info(DisjointSetMemberT *race_party);
#endif

/*
 * checking protype, called with gcc on cilk2c'ed Cilk program
 */
#ifdef CILK_NDL

extern const void *Cilk_internal_read_check(const void *p, size_t size, 
			 int linenum,char *filename,char *name);
extern const void *Cilk_internal_read_check4(const void *p, int linenum,
                         char *filename,char *name);
extern const void *Cilk_internal_read_check8(const void *p, int linenum,
                         char *filename, char *name);
extern const void *Cilk_internal_write_check(const void *p, size_t size, 
			 int op_type, int linenum, char *filename, char *name);
extern const void *Cilk_internal_write_check4(const void *p, int op_type,
			 int linenum,char *filename,char *name);
extern const void *Cilk_internal_write_check8(const void *p, int op_type,
			 int linenum,char *filename,char *name);
extern void Cilk_internal_spawnvar_check(const void *p, size_t size, 
	      int op_type,StackFrame *f,int linenum,char *filename,char *name);
extern void *Cilk_nd_malloc(size_t size, int linenum, char *filename, 
			    char *name);
extern void Cilk_nd_free(void *p, int linenum, char *filename, 
			 char *name);
extern void *Cilk_nd_memalign(size_t alignment, size_t size, int linenum, 
		       char *filename, char *name);
extern void *Cilk_nd_calloc(size_t nelem, size_t elsize, int linenum, 
			    char *filename, char *name);
extern void *Cilk_nd_realloc(void *s, size_t size, int linenum, char *filename,
		      char *name);
extern void *Cilk_nd_valloc(size_t size, int linenum, char *filename, 
			    char *name);
extern const void *Cilk_internal_free_check(const void *p, size_t size, 
			 int op_type, int linenum, char *filename, char *name);

#else

extern const void *Cilk_internal_read_check(const void *p, size_t size);
extern const void *Cilk_internal_read_check4(const void *p);
extern const void *Cilk_internal_read_check8(const void *p);
extern const void *Cilk_internal_write_check(const void *p, size_t size,
					     int op_type);
extern const void *Cilk_internal_write_check4(const void *p, int op_type);
extern const void *Cilk_internal_write_check8(const void *p, int op_type);
extern void Cilk_internal_spawnvar_check(const void *p, size_t size,
					 int op_type, StackFrame *f);
extern void *Cilk_nd_malloc(size_t size);
extern void Cilk_nd_free(void *p);
extern void *Cilk_nd_memalign(size_t alignment, size_t size);
extern void *Cilk_nd_calloc(size_t nelem, size_t elsize);
extern void *Cilk_nd_realloc(void *s, size_t size);
extern void *Cilk_nd_valloc(size_t size);
extern const void *Cilk_internal_free_check(const void *p, size_t size, 
					    int op_type);

#endif

/*
 * checking protype called in Cilk program, transformed in cilk2c.
 * no need to implement.
 */
#ifdef CILK_ND
extern void Cilk_read_check(void *p, size_t size);
extern void Cilk_write_check(void *p, size_t size);
#else
#define Cilk_read_check(p, size)
#define Cilk_write_check(p, size)
#endif
