/*
 * $Id: nondeterminator.c,v 1.1.2.14 1997/09/05 05:16:04 fengmd Exp $
 */

/*
 * $Log: nondeterminator.c,v $
 * Revision 1.1.2.14  1997/09/05 05:16:04  fengmd
 * FREE_ID changes to 2.  Fix cilk2c for Cilk_free var problem.
 *
 * Revision 1.1.2.13  1997/09/04 23:24:28  astark
 * Fixed typo in internal error message
 *
 * Revision 1.1.2.12  1997/09/04 23:17:47  astark
 * Fixed commutativity bug by introducing NO_INLET_ID
 *
 * Revision 1.1.2.11  1997/09/04 02:25:52  astark
 * Changed free race message
 *
 * Revision 1.1.2.10  1997/09/03 00:56:23  astark
 * Fixed race hashing to use filename and line num as well as var
 *
 * Revision 1.1.2.9  1997/08/29 07:38:40  fengmd
 * Merge malloc/free checker, race report hashing, garbage collection.
 *
 * Revision 1.1.2.8  1997/03/21 22:30:00  fengmd
 * USE ASSERT instead of CHECK. CHECK could not be masked by CILK_DEBUG.
 *
 * Revision 1.1.2.7  1997/03/13 18:48:09  fengmd
 * Declare pointer constant type to avoid gcc complaining.
 *
 * Revision 1.1.2.6  1997/03/04 00:07:01  fengmd
 * Fix static region to accommodate the region of global initialized variable.
 *
 * Revision 1.1.2.5  1997/03/03 21:02:14  fengmd
 * Add ND_VERBOSE to show the address range of shadow space.
 *
 * Revision 1.1.2.4  1997/02/18 21:10:05  fengmd
 * More message shown when rts error occurs.
 *
 * Revision 1.1.2.3  1997/02/18 19:56:35  fengmd
 * Add a command-line option for the Nondeterminator's heap size.
 *
 * Revision 1.1.2.2  1997/02/15 02:46:43  fengmd
 * Add Id and Log
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <cilk.h>
#include <cilk-lib.h>
#include <cilk-internal.h>
#include <cilk-cilk2c.h>

#ifdef CILK_ND

/***********************************************************\
 * region of shared memory
\***********************************************************/

DECLARE_SHARED_RO_VARIABLE(char*, heap_arena_start, );
DECLARE_SHARED_RO_VARIABLE(char*, heap_arena_end, );


/***********************************************************\
 * SP-bags algorithm
\***********************************************************/

DECLARE_SHARED_RO_VARIABLE(DisjointSetMemberT*, parent, );
DECLARE_SHARED_RO_VARIABLE(DisjointSetMemberT*, rank, );
DECLARE_SHARED_RO_VARIABLE(DisjointSetMemberT*, ref_count, );
DECLARE_SHARED_RO_VARIABLE(DisjointSetMemberT, next_proc_id, );
DECLARE_SHARED_RO_VARIABLE(DisjointSetMemberT, current_proc_id, );
DECLARE_SHARED_RO_VARIABLE(DisjointSetMemberT, current_inlet_id, );
DECLARE_SHARED_RO_VARIABLE(DisjointSetMemberT, collected_proc_id, );
DECLARE_SHARED_RO_VARIABLE(int, lreturn_size, );
DECLARE_SHARED_RO_VARIABLE(int, first_stop, );
DECLARE_SHARED_RO_VARIABLE(int, float_comm, );
DECLARE_SHARED_RO_VARIABLE(int, nd_arena, );

/* we need race_party in both versions to distinguish free/malloc bugs */
DECLARE_SHARED_RO_VARIABLE(DisjointSetMemberT*, race_party, );

/* variables needed to calculate shadow space offsets
   could be made into constants for efficiency */
DECLARE_SHARED_RO_VARIABLE(size_t, static_reader_offset, );
DECLARE_SHARED_RO_VARIABLE(size_t, static_writer_offset, );
DECLARE_SHARED_RO_VARIABLE(size_t, static_size, );
DECLARE_SHARED_RO_VARIABLE(size_t, heap_size, );
DECLARE_SHARED_RO_VARIABLE(size_t, both_plus_heap_size, );
DECLARE_SHARED_RO_VARIABLE(size_t, both_size2, );
DECLARE_SHARED_RO_VARIABLE(size_t, both_size4, );
extern int _edata;

WHEN_CILK_NDL(
DECLARE_SHARED_RO_VARIABLE(FileListT*, filelist, );
DECLARE_SHARED_RO_VARIABLE(int, race_linenum, );
DECLARE_SHARED_RO_VARIABLE(char*, race_filename, );
DECLARE_SHARED_RO_VARIABLE(char*, race_name, );
DECLARE_SHARED_RO_VARIABLE(RaceHashT*, race_table, [RACE_HASH_SIZE]);
)

static void report_race(const void *p);
static inline int is_commutative(unsigned int op1, unsigned int op2);
static inline int in_same_set(DisjointSetMemberT p1, DisjointSetMemberT p2);
WHEN_CILK_NDL(
static inline int find_filename(char *filename);
static inline char *get_filename(int index);
)

static DisjointSetMemberT alloc_proc_id();
static void collect_proc_id(DisjointSetMemberT i);

/* 
 * Disjoint-set operations 
 */
inline static DisjointSetMemberT make_set()
{
  register DisjointSetMemberT i;

  i = USE_SHARED_RO(current_proc_id);
  USE_SHARED_RO(current_inlet_id) = NO_INLET_ID;
  USE_SHARED_RO(parent)[i] = (DisjointSetMemberT) S_SET;
  USE_SHARED_RO(rank)[i] = 0;
  USE_SHARED_RO(ref_count)[i]++;
  return i;
}

inline static DisjointSetMemberT Union(DisjointSetMemberT x, 
                                       DisjointSetMemberT y)
{
  register DisjointSetMemberT rx, ry;

  rx = USE_SHARED_RO(rank)[x];
  ry = USE_SHARED_RO(rank)[y];
  if (rx > ry) {
    USE_SHARED_RO(parent)[y] = x;
    if (y >= FIRST_THREAD)
      USE_SHARED_RO(ref_count)[x]++;
    return x;
  }
  else {
    USE_SHARED_RO(parent)[x] = y;
    if (x >= FIRST_THREAD)
      USE_SHARED_RO(ref_count)[y]++; 
   if (rx == ry)
      USE_SHARED_RO(rank)[y]++;
    return y;
  }
}

inline static DisjointSetMemberT find_set(DisjointSetMemberT proc_id)
{
  register DisjointSetMemberT p = proc_id, pp, ppp;

  if (p > S_SET) {
    if ((p = USE_SHARED_RO(parent)[p]) > S_SET) {
      if ((p = USE_SHARED_RO(parent)[p]) > S_SET) {
	do {
	  pp = p;
	  p = USE_SHARED_RO(parent)[p];
	} while (p > S_SET);
	do {
	  ppp = USE_SHARED_RO(parent)[proc_id];
	  if (USE_SHARED_RO(ref_count)[proc_id]) {
	    USE_SHARED_RO(parent)[proc_id] = pp;
	    USE_SHARED_RO(ref_count)[pp]++;
	    USE_SHARED_RO(ref_count)[ppp]--;
	  }
	  else
	    collect_proc_id(proc_id);
	  proc_id = ppp;
	} while (proc_id != pp);
      }
    }
  }
  return p;
}

/*
 * allocate a procedure ID to a new spawned procedure
 * Cilk_alloc_proc_id_spbags is called in INIT_FRAME before frame is allocated
 */
inline void Cilk_alloc_proc_id_spbags(StackFrame *f)
{
  USE_SHARED_RO(current_proc_id) = alloc_proc_id();
}

/*
 * initialize the SP-bag per instance of the procedure.
 * Cilk_init_spbags is called in INIT_FRAME when a procedure starts
 */
inline void Cilk_init_spbags(StackFrame *f)
{
  f->proc_id = f->s_set = make_set();
  f->p_set = P_SET;
  f->inlet_id = NO_INLET_ID;
}

/*
 * union the S-bag of the child into the P-bag of the parent.
 * Cilk_lreturn_spbags is called in _XPOP_FRAME_RESULT or _XPOP_FRAME_NORESULT
 */
inline void Cilk_lreturn_spbags(StackFrame *f, volatile StackFrame *child)
{
  USE_SHARED_RO(parent)[f->p_set = Union(f->p_set, child->s_set)] = P_SET;
  USE_SHARED_RO(current_proc_id) = f->proc_id;
  USE_SHARED_RO(current_inlet_id) = f->inlet_id;
  if (((--USE_SHARED_RO(ref_count)[child->proc_id]) == 0) &&
      (f->p_set != child->proc_id)) {
    USE_SHARED_RO(ref_count)[USE_SHARED_RO(parent)[child->proc_id]]--;
    collect_proc_id(child->proc_id);
  }
}

/*
 * union the P-bag of the current procedure into its S-bag.
 * Cilk_sync_spbags is called in _AT_SYNC_FAST
 */
inline void Cilk_sync_spbags(StackFrame *f)
{
  USE_SHARED_RO(parent)[f->s_set = Union(f->p_set, f->s_set)] = S_SET;
  f->p_set = P_SET;
  f->inlet_id = NO_INLET_ID; /* after sync, a new inlet id is needed */
}

/*
 * routine to check a read access of an address p with p's reader.
 * p's reader may be updated.
 */
inline static int read_check_reader(DisjointSetMemberT *readerp,
				    DisjointSetMemberT my_proc_id)
{
  register DisjointSetMemberT reader = *readerp;

  if (find_set(reader) == S_SET)  {
    update_ref_count(readerp, reader, my_proc_id, my_proc_id);
    WHEN_CILK_NDL(save_shadow_info(readerp);)
    return 0;
  }
  return 1;
}

/*
 * routine to check a read access of an address p with p's writer.
 */
inline static void read_check_writer(const void *p,DisjointSetMemberT *writerp)
{
  if (find_set(PROC_ID(*writerp)) == P_SET)  {
    USE_SHARED_RO(race_party) = writerp;
    report_race(p);
  }
}

/*
 * routine to check a write access of an address p with p's reader.
 * p's writer and reader may also be updated.
 */
inline static void write_check(const void *p, DisjointSetMemberT *readerp,
			       DisjointSetMemberT *writerp,
			       DisjointSetMemberT my_proc_id, int op_type)
{
  register DisjointSetMemberT reader = *readerp;

  if (find_set(reader) == P_SET) {
    register DisjointSetMemberT writer = *writerp;

    if (!((PROC_ID(writer) == USE_SHARED_RO(current_inlet_id)) &&
	  is_commutative(OP_TYPE(writer), op_type))) {
      USE_SHARED_RO(race_party) = readerp;

      /* a write to an uninitialized location is not an error */
      if (my_proc_id == FREE_ID) fprintf(stderr, "Free: ");

      report_race(p);
    }
  }
  else {
    update_ref_count(readerp, reader, my_proc_id, my_proc_id);
    update_ref_count(writerp, PROC_ID(*writerp), my_proc_id, 
		     OP_TYPE_PROC_ID(my_proc_id, op_type));
    WHEN_CILK_NDL(save_shadow_info(readerp);)
    WHEN_CILK_NDL(save_shadow_info(writerp);)
  }
}

/*
 * check the read access in region [p, p+size)
 */
#ifdef CILK_NDL
const void *Cilk_internal_read_check(const void *p, size_t size, int linenum,
				     char *filename, char *name)
#else
const void *Cilk_internal_read_check(const void *p, size_t size)
#endif
{
  register int *pp;
  register DisjointSetMemberT my_proc_id = USE_SHARED_RO(current_proc_id);

  WHEN_CILK_NDL(USE_SHARED_RO(race_linenum) = linenum; 
		USE_SHARED_RO(race_filename) = filename;
		USE_SHARED_RO(race_name) = name;)

  size = (((size + sizeof(int) - 1) & (~(sizeof(int) -1))) / sizeof(int));
  pp = (int *) (((unsigned int) p) & (~(sizeof(int) -1)));
      
  if (pp >= (int *) USE_SHARED_RO(heap_arena_start)) {
    if (pp+size-1 < (int *) USE_SHARED_RO(heap_arena_end)) {
      for (; size; size -= 1, pp++) {
	if (read_check_reader(HEAP_READER(pp), my_proc_id))
	  read_check_writer(pp, HEAP_WRITER(pp));
      }
      return p;
    }
  }
      
  if (pp >= (int *) STATIC_ARENA_START) {
    if (pp+size-1 < (int *) STATIC_ARENA_END) {
      for (; size; size -= 1, pp++) {
	if (read_check_reader(STATIC_READER(pp), my_proc_id))
	  read_check_writer(pp, STATIC_WRITER(pp));
      }
      return p;
    }
  }
  
  return p;

}

/*
 * check the read access in region [p, p+4).
 * this is to optimize the common case of "read an integer or float".
 */
#ifdef CILK_NDL
const void *Cilk_internal_read_check4(const void *p, int linenum, 
                                      char *filename, char *name)
#else
const void *Cilk_internal_read_check4(const void *p)
#endif
{
  register DisjointSetMemberT my_proc_id = USE_SHARED_RO(current_proc_id);

  WHEN_CILK_NDL(USE_SHARED_RO(race_linenum) = linenum; 
		USE_SHARED_RO(race_filename) = filename;
		USE_SHARED_RO(race_name) = name;)
  
  if (p >= (void *) USE_SHARED_RO(heap_arena_start)) {
    if (p < (void *) USE_SHARED_RO(heap_arena_end)) {
      if (read_check_reader(HEAP_READER(p), my_proc_id))
	read_check_writer(p, HEAP_WRITER(p));
      return p;
    }
  }
      
  if (p >= (void *) STATIC_ARENA_START) {
    if (p < (void *) STATIC_ARENA_END) {
      if (read_check_reader(STATIC_READER(p), my_proc_id))
	read_check_writer(p, STATIC_WRITER(p));
      return p;
    }
  }
      
  return p;
}

/*
 * check the read access in region [p, p+8).
 * this is to optimize the common case of "read a double".
 */
#ifdef CILK_NDL
const void *Cilk_internal_read_check8(const void *p, int linenum, 
                                      char *filename, char *name)
#else
const void *Cilk_internal_read_check8(const void *p)
#endif
{
  register const void *pp;
  register DisjointSetMemberT my_proc_id = USE_SHARED_RO(current_proc_id);
      
  WHEN_CILK_NDL(USE_SHARED_RO(race_linenum) = linenum; 
		USE_SHARED_RO(race_filename) = filename;
		USE_SHARED_RO(race_name) = name;)

  if (p >= (void *) USE_SHARED_RO(heap_arena_start)) {
    if (p+4 < (void *) USE_SHARED_RO(heap_arena_end)) {
      if (read_check_reader(HEAP_READER(p), my_proc_id))
	read_check_writer(p, HEAP_WRITER(p));
      
      pp = p + 4;
      if (read_check_reader(HEAP_READER(pp), my_proc_id))
	read_check_writer(pp, HEAP_WRITER(pp));
      return p;
    }
  }

  if (p >= (void *) STATIC_ARENA_START) {
    if (p+4 < (void *) STATIC_ARENA_END) {
      if (read_check_reader(STATIC_READER(p), my_proc_id))
	read_check_writer(p, STATIC_WRITER(p));
      
      pp = p + 4;
      if (read_check_reader(STATIC_READER(pp), my_proc_id))
	read_check_writer(pp, STATIC_WRITER(pp));
      return p;
    }
  }

  return p;
}

/*
 * check the write access in region [p, p+size)
 */
#ifdef CILK_NDL
const void *Cilk_internal_write_check(const void *p, size_t size, int op_type, 
				      int linenum, char *filename, char *name)
#else
const void *Cilk_internal_write_check(const void *p, size_t size, int op_type)
#endif
{
  register int *pp;
  register DisjointSetMemberT my_proc_id = USE_SHARED_RO(current_proc_id);

  WHEN_CILK_NDL(USE_SHARED_RO(race_linenum) = linenum; 
		USE_SHARED_RO(race_filename) = filename;
		USE_SHARED_RO(race_name) = name;)

  size = (((size + sizeof(int) - 1) & (~(sizeof(int) -1))) / sizeof(int));
  pp = (int *) (((unsigned int) p) & (~(sizeof(int) -1)));

  if (pp >= (int *) USE_SHARED_RO(heap_arena_start)) {
    if (pp+size-1 < (int *) USE_SHARED_RO(heap_arena_end)) {
      for (; size; size -= 1, pp++) {
	write_check(pp, HEAP_READER(pp), HEAP_WRITER(pp), my_proc_id, op_type);
      }
      return p;
    }
  }
  
  if (pp >= (int *) STATIC_ARENA_START) {
    if (pp+size-1 < (int *) STATIC_ARENA_END) {
      for (; size; size -= 1, pp++) {
	write_check(pp, STATIC_READER(pp), STATIC_WRITER(pp), my_proc_id, op_type);
      }
      return p;
    }
  }

  return p;
}

/*
 * check the write access in region [p, p+4).
 * this is to optimize the common case of "write an integer or float".
 */
#ifdef CILK_NDL
const void *Cilk_internal_write_check4(const void *p, int op_type, int linenum,
				       char *filename, char *name)
#else
const void *Cilk_internal_write_check4(const void *p, int op_type)
#endif
{
  register DisjointSetMemberT my_proc_id = USE_SHARED_RO(current_proc_id);

  WHEN_CILK_NDL(USE_SHARED_RO(race_linenum) = linenum; 
		USE_SHARED_RO(race_filename) = filename;
		USE_SHARED_RO(race_name) = name;)

  if (p >= (void *) USE_SHARED_RO(heap_arena_start)) {
    if (p < (void *) USE_SHARED_RO(heap_arena_end)) {
      write_check(p, HEAP_READER(p), HEAP_WRITER(p), my_proc_id, op_type);
      return p;
    }
  }

  if (p >= (void *) STATIC_ARENA_START) {
    if (p < (void *) STATIC_ARENA_END) {
      write_check(p, STATIC_READER(p), STATIC_WRITER(p), my_proc_id, op_type);
      return p;
    }
  }

  return p;
}

/*
 * check the write access in region [p, p+8).
 * this is to optimize the common case of "write a double".
 */
#ifdef CILK_NDL
const void *Cilk_internal_write_check8(const void *p, int op_type, int linenum,
				       char *filename, char *name)
#else
const void *Cilk_internal_write_check8(const void *p, int op_type)
#endif
{
  register const void *pp;
  register DisjointSetMemberT my_proc_id = USE_SHARED_RO(current_proc_id);

  WHEN_CILK_NDL(USE_SHARED_RO(race_linenum) = linenum; 
		USE_SHARED_RO(race_filename) = filename;
		USE_SHARED_RO(race_name) = name;)

  if (p >= (void *) USE_SHARED_RO(heap_arena_start)) {
    if (p+4 < (void *) USE_SHARED_RO(heap_arena_end)) {
      write_check(p, HEAP_READER(p), HEAP_WRITER(p), my_proc_id, op_type);

      pp = p + 4;
      write_check(pp, HEAP_READER(pp), HEAP_WRITER(pp), my_proc_id, op_type);
      return p;
    }
  }

  if (p >= (void *) STATIC_ARENA_START) {
    if (p+4 < (void *) STATIC_ARENA_END) {
      write_check(p, STATIC_READER(p), STATIC_WRITER(p), my_proc_id, op_type);
      
      pp = p + 4;
      write_check(pp, STATIC_READER(pp), STATIC_WRITER(pp), my_proc_id, op_type);
      return p;
    }
  }

  return p;
}

/* 
 * similar to Cilk_internal_write_check, except using FREE_ID as 
 * current_proc_id 
 */
#ifdef CILK_NDL
const void *Cilk_internal_free_check(const void *p, size_t size, int op_type, 
				      int linenum, char *filename, char *name)
#else
const void *Cilk_internal_free_check(const void *p, size_t size, int op_type)
#endif
{
  register int *pp;
  
  WHEN_CILK_NDL(USE_SHARED_RO(race_linenum) = linenum; 
		USE_SHARED_RO(race_filename) = filename;
		USE_SHARED_RO(race_name) = name;)

  size = (((size + sizeof(int) - 1) & (~(sizeof(int) -1))) / sizeof(int));
  pp = (int *) (((unsigned int) p) & (~(sizeof(int) -1)));

  if (pp >= (int *) USE_SHARED_RO(heap_arena_start)) {
    if (pp+size-1 < (int *) USE_SHARED_RO(heap_arena_end)) {
      for (; size; size -= 1, pp++) {
	/*
	 * if the shadow space contains FREE_ID, that is not a race since 
	 * that represents the extra memory returned as part of the malloc 
	 * bucket 
	 */
	if (*HEAP_READER(pp) == FREE_ID) break;
	write_check(pp, HEAP_READER(pp), HEAP_WRITER(pp), FREE_ID, op_type);
	}
      return p;
    }
  }
  
  if (pp >= (int *) STATIC_ARENA_START) {
    if (pp+size-1 < (int *) STATIC_ARENA_END) {
      for (; size; size -= 1, pp++) {
	write_check(pp, STATIC_READER(pp), STATIC_WRITER(pp), FREE_ID, op_type);
      }
      return p;
    }
  }

  return p;
}
    
inline static DisjointSetMemberT make_inlet_id(StackFrame *f)
{
  register DisjointSetMemberT i;

  i = f->inlet_id = alloc_proc_id();
  USE_SHARED_RO(rank)[i] = 0;
  USE_SHARED_RO(parent)[f->p_set = Union(f->p_set, i)] = P_SET;
  return i;
}

inline void spawnvar_check(const void *p, DisjointSetMemberT *readerp,
			   DisjointSetMemberT *writerp,
			   DisjointSetMemberT my_proc_id, int op_type,
			   StackFrame *f)
{
  register DisjointSetMemberT reader = *readerp;
  register DisjointSetMemberT writer = *writerp;

  if (find_set(reader) == P_SET) {
    if (!((PROC_ID(writer) == f->inlet_id) &&
	  is_commutative(OP_TYPE(writer), op_type))) {
      USE_SHARED_RO(race_party) = readerp;
      report_race(p);
    }
  }
  else {
    if (!in_same_set(reader, my_proc_id)) {
      register DisjointSetMemberT inlet_id;

      inlet_id = (f->inlet_id == NO_INLET_ID) ? make_inlet_id(f) : f->inlet_id;
      update_ref_count(readerp, reader, inlet_id, inlet_id);
      update_ref_count(writerp, PROC_ID(writer), inlet_id, 
		       OP_TYPE_PROC_ID(inlet_id, op_type));
      WHEN_CILK_NDL(save_shadow_info(readerp);)
      WHEN_CILK_NDL(save_shadow_info(writerp);)
      }
    else if (!in_same_set(PROC_ID(writer), my_proc_id)) {
      register DisjointSetMemberT inlet_id;

      inlet_id = (f->inlet_id == NO_INLET_ID) ? make_inlet_id(f) : f->inlet_id;
      update_ref_count(writerp, PROC_ID(writer), inlet_id, 
		       OP_TYPE_PROC_ID(inlet_id, op_type));
      WHEN_CILK_NDL(save_shadow_info(writerp);)
      }
  }
}

/*
 * check commutative of op_type in the shadow space
 */
#ifdef CILK_NDL
void Cilk_internal_spawnvar_check(const void *p, size_t size, int op_type,
		        StackFrame *f, int linenum, char *filename, char *name)
#else
void Cilk_internal_spawnvar_check(const void *p, size_t size, int op_type,
				  StackFrame *f)
#endif
{
  register int *pp;
  register DisjointSetMemberT my_proc_id = USE_SHARED_RO(current_proc_id);

  WHEN_CILK_NDL(USE_SHARED_RO(race_linenum) = linenum; 
		USE_SHARED_RO(race_filename) = filename;
		USE_SHARED_RO(race_name) = name;)

  size = (((size + sizeof(int) - 1) & (~(sizeof(int) -1))) / sizeof(int));
  pp = (int *) (((unsigned int) p) & (~(sizeof(int) -1)));

  if (pp >= (int *) USE_SHARED_RO(heap_arena_start)) {
    if (pp+size-1 < (int *) USE_SHARED_RO(heap_arena_end)) {
      for (; size; size -= 1, pp++) {
        spawnvar_check(pp, HEAP_READER(pp), HEAP_WRITER(pp), my_proc_id, 
		       op_type, f);
      }
      return;
    }
  }

  if (pp >= (int *) STATIC_ARENA_START) {
    if (pp+size-1 < (int *) STATIC_ARENA_END) {
      for (; size; size -= 1, pp++) {
        spawnvar_check(pp, STATIC_READER(pp), STATIC_WRITER(pp), my_proc_id, 
		       op_type, f);
      }
      return;
    }
  }

  CILK_CHECK(pp, ("spawn receiver address %p out of range\n", pp));
  return;
}

/* 
 * hash functions allow us to report races only once, using the position 
 * in the user's code to identify the race.
 */
#ifdef CILK_NDL
static int race_hash()
{
  int index;
  RaceHashT *race, **prace;
  char *shadow_name;
  int shadow_info, current_info;

  shadow_info = *SHADOW_INFO(USE_SHARED_RO(race_party));
  shadow_name = *SHADOW_VAR(USE_SHARED_RO(race_party));
  current_info = (USE_SHARED_RO(race_linenum) << 8) +
                 find_filename(USE_SHARED_RO(race_filename));
  index = ((((int) USE_SHARED_RO(race_name) + (int) shadow_name ) >> 2) +
	   current_info + shadow_info) % RACE_HASH_SIZE;
  prace = USE_SHARED_RO(race_table) + index;
  while ((race = *prace) != NULL) {
    if (((race->current_name == USE_SHARED_RO(race_name)) &&
	 (race->shadow_name == shadow_name) &&
	 (race->current_info == current_info) &&
	 (race->shadow_info == shadow_info)) ||
	((race->current_name == shadow_name) &&
	 (race->shadow_name == USE_SHARED_RO(race_name)) &&
	 (race->current_info == shadow_info) &&
	 (race->shadow_info == current_info)))
      return 0;
    prace = &(race->next);
  }
  *prace = (RaceHashT *)Cilk_malloc(sizeof(RaceHashT));
  (*prace)->current_name = USE_SHARED_RO(race_name);
  (*prace)->current_info = current_info;
  (*prace)->shadow_name = shadow_name;
  (*prace)->shadow_info = shadow_info;
  (*prace)->next = NULL;
  return 1;
}

static void report_full_race(const void *p)
{
    fprintf(stderr, ": `%s'(%s:%d) with `%s'(%s:%d)",
      *(SHADOW_VAR(USE_SHARED_RO(race_party))),
      get_filename(SHADOW_INFO_FILE(SHADOW_INFO(USE_SHARED_RO(race_party)))),
      SHADOW_INFO_LINE(SHADOW_INFO(USE_SHARED_RO(race_party))),
      USE_SHARED_RO(race_name),
      USE_SHARED_RO(race_filename), USE_SHARED_RO(race_linenum));
}

static void ndl_report_serial_bug()
{
  fprintf(stderr, ": `%s' (%s:%d)", USE_SHARED_RO(race_name),
	  USE_SHARED_RO(race_filename), USE_SHARED_RO(race_linenum));
}

static void report_race(const void *p)
{
  int hashfound;
  
  if (*(USE_SHARED_RO(race_party)) == FREE_ID) {
    /* access to freed memory -- could be a race or a serial bug */
    if (SHADOW_INFO_FILE(SHADOW_INFO(USE_SHARED_RO(race_party))) 
	== INIT_INDEX) {
      /* set the shadow var to NULL for the hash table entry */
      *SHADOW_VAR(USE_SHARED_RO(race_party)) = NULL;
      if ((hashfound = race_hash())) {
	fprintf(stderr, "Access to non-allocated memory @ 0x%p", p);
	ndl_report_serial_bug();
      }
    }
    else {
      /* a free/access race */
      if ((hashfound = race_hash())) {
	fprintf(stderr, "Free: Determinacy race @ 0x%p", p); 
	report_full_race(p);
      }
    }
  }
  else { /* ordinary determinacy race case */
    if ((hashfound = race_hash())) {
      fprintf(stderr, "Determinacy race @ 0x%p", p);
      report_full_race(p);
    }
  }

  if (hashfound) fprintf(stderr, "\n");
  if (USE_SHARED_RO(first_stop)) 
    Cilk_die("The first determinacy race is found, and the program is terminated\n");
}

#else /* not CILK_NDL */

static void report_race(const void *p)
{
   if (*(USE_SHARED_RO(race_party)) == FREE_ID) {
    /* access to freed memory -- could be a race or a serial bug */
     fprintf(stderr, "Access to non-allocated memory @ 0x%p", p);
   }
   else
     { /* ordinary determinacy race case */
       fprintf(stderr, "Determinacy race @ 0x%p", p);
     }

   fprintf(stderr, "\n");
   if (USE_SHARED_RO(first_stop)) 
     Cilk_die("The first determinacy race is found, and the program is terminated\n");
}
#endif /* CILK_NDL */

/* 
 * test whether two operations are communitive or not 
 */
static inline int is_commutative(unsigned int op1, unsigned int op2)
{
  register unsigned int operator1, operator2, type1, type2;

  operator1 = CHECK_OPERATOR(op1);
  operator2 = CHECK_OPERATOR(op2);
  type1 = CHECK_TYPE(op1);
  type2 = CHECK_TYPE(op2);

  CILK_ASSERT(type1 == type2);
  if (operator1 == operator2) {
    switch (type1) {
    case INTtype:
      return ((operator1 <= ANDassign) && (operator1 != DIVassign)) ? 1 : 0;
    case FLOATtype:
      return USE_SHARED_RO(float_comm) ? 
               ((operator1 <= DIVassign) ? 1 : 0) : 0;
    case POINTtype:
      return (operator1 <= PLUSassign) ? 1 : 0;
    case OTHERtype:
      return 0;
    default:
      CILK_CHECK(type1 > OTHERtype, ("wrong type in is_commutative\n"));
      return 0;
    }
  }
  else
    return 0;
}

/*
 * test whether two procedures are in the same set
 */
static inline int in_same_set(DisjointSetMemberT p1, DisjointSetMemberT p2)
{
  register DisjointSetMemberT pp1, pp2;

  if (p1 > S_SET) {
    for (pp1 = USE_SHARED_RO(parent)[p1]; 
	 pp1 > S_SET; 
	 p1 = pp1, pp1 = USE_SHARED_RO(parent)[pp1]);
  }
  if (p2 > S_SET) {
    for (pp2 = USE_SHARED_RO(parent)[p2]; 
	 pp2 > S_SET; 
	 p2 = pp2, pp2 = USE_SHARED_RO(parent)[pp2]);
  }
  return (p1 == p2);
}

/* initialize private variables */
static void init_private()
{
  USE_SHARED_RO(next_proc_id) = FIRST_THREAD;
  USE_SHARED_RO(collected_proc_id) = 0;
  WHEN_CILK_NDL(USE_SHARED_RO(filelist) = NULL;)
  WHEN_CILK_NDL(memset(USE_SHARED_RO(race_table), 0, RACE_HASH_SIZE*sizeof(RaceHashT *));)
}

/* malloc space for parent and rank used in disjoint-set */
static void init_disjoint_set()
{
  /* if the nd_arena is set too large, just use as much as we need */
  if (USE_SHARED_RO(nd_arena) > MAX_PROC_NUMBER) {
    USE_SHARED_RO(nd_arena) = MAX_PROC_NUMBER;
#if ND_VERBOSE
    printf("Using maximum useful nd_heap size, %ld bytes\n", MAX_PROC_NUMBER);
#endif
  }
  
  if ((USE_SHARED_RO(parent) =
      (DisjointSetMemberT *)malloc(USE_SHARED_RO(nd_arena)*
				   sizeof(DisjointSetMemberT))) == NULL)
    Cilk_die("malloc error in init_disjoint_set for parent\n"
	     "Please run with smaller Nondeterminator heap size (by -nd-heap option)\n");

  if ((USE_SHARED_RO(rank) = 
      (DisjointSetMemberT *)malloc(USE_SHARED_RO(nd_arena)*
				   sizeof(DisjointSetMemberT))) == NULL)
    Cilk_die("malloc error in init_disjoint_set for rank\n"
	     "Please run with smaller Nondeterminator heap size (by -nd-heap option)\n");

  if ((USE_SHARED_RO(ref_count) = 
      (DisjointSetMemberT *)malloc(USE_SHARED_RO(nd_arena)*
				   sizeof(DisjointSetMemberT))) == NULL)
    Cilk_die("malloc error in init_disjoint_set for ref_count\n"
	     "Please run with smaller Nondeterminator heap size (by -nd-heap option)\n");

  /* put the FREE_ID into a separate P-bag */
  USE_SHARED_RO(parent)[FREE_ID] = P_SET;
  USE_SHARED_RO(rank)[FREE_ID] = 0;
  USE_SHARED_RO(ref_count)[FREE_ID] = (USE_SHARED_RO(heap_size) >> 1) + 1;

  USE_SHARED_RO(ref_count)[S_SET] = (USE_SHARED_RO(static_size) >> 1) + 1;
}

/* set up the reader and writer shadow space */
static void init_shadow_space()
{
  void *p;
  size_t static_size, heap_size;
  char *static_reader_fname = "/tmp/_Cilk_static_reader";
  int static_reader_fd;
  void *start_static_read, *start_static_write;
  void *start_heap_read, *start_heap_write;
  size_t start_static_arena, start_heap_arena, page_size;
  size_t both_size;
  int junk = 42;
  size_t i;
  WHEN_CILK_NDL(
  char *static_reader_info_fname = "/tmp/_Cilk_static_reader_info";
  int static_reader_info_fd;
  )

  /* align the starting address of static and heap regions */
  page_size = Cilk_get_page_size();
  start_static_arena = ((size_t) STATIC_ARENA_START) / page_size * page_size;
  start_heap_arena = ((size_t) USE_SHARED_RO(heap_arena_start)) /
                     page_size * page_size;

  /* get the size of static and heap regions, round to page size */
  static_size = (size_t) STATIC_ARENA_END - start_static_arena;
  static_size = (size_t) ((static_size / page_size * page_size) +
                          (((static_size % page_size) == 0) ? 0:page_size));
  heap_size = (size_t) USE_SHARED_RO(heap_arena_end) - start_heap_arena;
  heap_size = (size_t) ((heap_size / page_size * page_size) +
                          (((heap_size % page_size) == 0) ? 0:page_size));

  USE_SHARED_RO(static_size) = static_size;
  USE_SHARED_RO(heap_size) = heap_size;
  both_size = heap_size + static_size;
  USE_SHARED_RO(both_plus_heap_size) = both_size + heap_size; 
                                           /* for HEAP_READER macro */
  USE_SHARED_RO(both_size2) = 2*both_size; /* for SHADOW_INFO macros */
  USE_SHARED_RO(both_size4) = 4*both_size; /* for SHADOW_VAR macros */
  USE_SHARED_RO(static_reader_offset) =
     start_heap_arena - USE_SHARED_RO(both_size2) - start_static_arena;
  USE_SHARED_RO(static_writer_offset) =
     start_heap_arena - both_size - start_static_arena;

  /* get starting addresses of shadow spaces */
  start_static_read = STATIC_READER(start_static_arena);
  start_static_write = STATIC_WRITER(start_static_arena);
  start_heap_read = HEAP_READER(start_heap_arena);
  start_heap_write = HEAP_WRITER(start_heap_arena);

#if ND_VERBOSE
  printf("static 0x%p (align to 0x%x) --- 0x%p\n", 
	 STATIC_ARENA_START, start_static_arena, STATIC_ARENA_END);
  printf("    reader 0x%p", start_static_read);
  WHEN_CILK_NDL(printf(", reader info 0x%p, reader var 0x%p",
		SHADOW_INFO(start_static_read), 
		SHADOW_VAR(start_static_read));)
  printf("\n    writer 0x%p", start_static_write);
  WHEN_CILK_NDL(printf(", writer info 0x%p, writer var 0x%p",
		SHADOW_INFO(start_static_write),
		SHADOW_VAR(start_static_write));)
  printf("\nheap 0x%p (align to 0x%x) --- 0x%p\n",
	 USE_SHARED_RO(heap_arena_start), start_heap_arena,
	 USE_SHARED_RO(heap_arena_end));
  printf("    reader 0x%p", start_heap_read);
  WHEN_CILK_NDL(printf(", reader info 0x%p, reader var 0x%p",
		SHADOW_INFO(start_heap_read), 
		SHADOW_VAR(start_heap_read));)
  printf("\n    writer 0x%p", start_heap_write);
  WHEN_CILK_NDL(printf(", writer info 0x%p, writer var 0x%p",
		SHADOW_INFO(start_heap_write),
		SHADOW_VAR(start_heap_write));)
  printf("\n");
  printf("End of data segment 0x%p\n", &_edata);
#endif	       

  /* open files for mmap */
  static_reader_fd = open(static_reader_fname, O_RDWR | O_CREAT, 0600);
  CILK_CHECK(static_reader_fd >= 0, ("Open %s error\n", static_reader_fname));
  WHEN_CILK_NDL(
  static_reader_info_fd = open(static_reader_info_fname, O_RDWR | O_CREAT, 
			       0600);
  CILK_CHECK(static_reader_info_fd >= 0, 
	     ("Open %s error\n", static_reader_info_fname));
  )

  /* delete temporary file */
  unlink(static_reader_fname);
  WHEN_CILK_NDL(unlink(static_reader_info_fname);)

  CILK_CHECK(lseek(static_reader_fd, 
		   USE_SHARED_RO(both_size2) - sizeof(int), SEEK_SET) ==
	           USE_SHARED_RO(both_size2) - sizeof(int),
	     ("error in init_shadow_space to lseek static_reader_fd!\n"
	      "Please run with smaller heap size (by -heap option)\n"));
  CILK_CHECK(write(static_reader_fd, &junk, sizeof(int)) == sizeof(int),
	     ("error in init_shadow_space to write static_reader_fd!\n"
	      "Please run with smaller heap size (by -heap option)\n"));

  WHEN_CILK_NDL(
  CILK_CHECK(lseek(static_reader_info_fd, 
		   USE_SHARED_RO(both_size4) - sizeof(int), SEEK_SET) ==
	           USE_SHARED_RO(both_size4) - sizeof(int),
	     ("error in init_shadow_space to lseek static_reader_info_fd!\n"
	      "Please run with smaller heap size (by -heap option)\n"));
  CILK_CHECK(write(static_reader_info_fd, &junk, sizeof(int)) == sizeof(int),
	     ("error in init_shadow_space to write static_reader_info_fd!\n"
	      "Please run with smaller heap size (by -heap option)\n"));
  )


  /* minimal check for overlap */
  CILK_CHECK(((int *)start_static_read > (int *)&_edata),
	     ("Total space required too large\n"
	      "Please run with smaller heap size (by -heap option)\n"));
  WHEN_CILK_NDL(
    CILK_CHECK(((int *) SHADOW_VAR(start_static_read) > &_edata),
	       ("Total space required too large\n"
		"Please run with smaller heap size (by -heap option)\n"));
  )

  /* this represents the order we expect */
  /* SRV | HRV | SWV | HWV | SRI | HRI | SWI | HWI | SR | HR | SW | HW | H */
  CILK_ASSERT(SHADOW_VAR(start_static_read) < SHADOW_VAR(start_heap_read));
  CILK_ASSERT(SHADOW_VAR(start_heap_read) < SHADOW_VAR(start_static_write));
  CILK_ASSERT(SHADOW_VAR(start_static_write) < SHADOW_VAR(start_heap_write));
  CILK_ASSERT((void *)SHADOW_VAR(start_heap_write) <
	      (void *)SHADOW_INFO(start_static_read));
  CILK_ASSERT(SHADOW_INFO(start_static_read) < SHADOW_INFO(start_heap_read));
  CILK_ASSERT(SHADOW_INFO(start_heap_read) < SHADOW_INFO(start_static_write));
  CILK_ASSERT(SHADOW_INFO(start_static_write) < SHADOW_INFO(start_heap_write));
  CILK_ASSERT((void *)SHADOW_INFO(start_heap_write) < start_static_read);
  CILK_ASSERT(start_static_read < start_heap_read);
  CILK_ASSERT(start_heap_read < start_static_write);
  CILK_ASSERT(start_static_write < start_heap_write);
  /* CILK_ASSERT(start_heap_write < (void *)start_heap_arena); */

  /* now do mmap */
  p = mmap(start_static_read, 2*both_size, PROT_READ | PROT_WRITE, 
	   MAP_SHARED | MAP_FIXED, static_reader_fd, 0);
  if (p == MAP_FAILED) {
    perror("mmap");
    Cilk_die("mmap error in init_shadow_space for reader/writer shadow spaces\n");
  }
  
  WHEN_CILK_NDL(
  p = mmap((void *)SHADOW_VAR(start_static_read), 4*both_size, 
	   PROT_READ | PROT_WRITE, MAP_SHARED | MAP_FIXED, 
	   static_reader_info_fd, 0);
  if (p == MAP_FAILED)
    Cilk_die("mmap error in init_shadow_space for reader/writer info shadow spaces\n");
  )
     
  for (i=0;i< (static_size>>2) ;i++) {
    *((DisjointSetMemberT *)start_static_read + i) = S_SET;
    *((DisjointSetMemberT *)start_static_write + i) = S_SET;
  }
}

void nondeterminator_init()
{
  CILK_CHECK(USE_SHARED_RO(active_size) == 1,
	     ("The Nondeterminator runs only on one processor\n"));

  CILK_CHECK(PROC_ID_BIT + OPERATOR_BIT + TYPE_BIT <= 32,
	     ("Operator type for each shadow location need at least 6 bit"));

  init_private();
  init_shadow_space();
  init_disjoint_set();
}
     
WHEN_CILK_NDL(
inline void save_shadow_info(DisjointSetMemberT *race_party)
{
  *(SHADOW_INFO(race_party)) = (USE_SHARED_RO(race_linenum) << 8) +
                               find_filename(USE_SHARED_RO(race_filename));
  *(SHADOW_VAR(race_party)) = USE_SHARED_RO(race_name);
}

static inline int find_filename(char *filename)
{
  FileListT *fl;
  int index;

  for (fl = USE_SHARED_RO(filelist), index = START_FILE_INDEX; 
       fl; 
       fl = fl->next, index++) {
    if (filename == fl->filename)
      return fl->index;
  }
  fl = (FileListT *) malloc(sizeof(FileListT));
  fl->next = USE_SHARED_RO(filelist);
  fl->filename = filename;
  fl->index = index;
  USE_SHARED_RO(filelist) = fl;
  return index;
}

static inline char *get_filename(int index)
{
  FileListT *fl;

  for (fl = USE_SHARED_RO(filelist); fl && (fl->index != index); fl = fl->next);
  CILK_ASSERT(fl);
  return fl->filename;
}
)

/*
 * allocate a procedure ID from the garbage-collected pool, or generated a new
 */
static DisjointSetMemberT alloc_proc_id()
{
  DisjointSetMemberT i;

  if (USE_SHARED_RO(collected_proc_id)) {
    i = USE_SHARED_RO(collected_proc_id);
    USE_SHARED_RO(collected_proc_id) = USE_SHARED_RO(parent)[i];
  }
  else {
    i = USE_SHARED_RO(next_proc_id);
    CILK_CHECK(i < USE_SHARED_RO(nd_arena),
	       ("The Nondeterminator runs out %d bytes of disjoint-set space\n"
		"Please try a larger Nondeterminator heap size (by -nd-heap option)\n",
		USE_SHARED_RO(nd_arena)));
    USE_SHARED_RO(next_proc_id)++;
    USE_SHARED_RO(ref_count)[i] = 0;
  }
  return i;
}

/*
 * garbage-collect a procedure ID
 */
static void collect_proc_id(DisjointSetMemberT i)
{
  CILK_ASSERT(i >= FIRST_THREAD);
  USE_SHARED_RO(parent)[i] = USE_SHARED_RO(collected_proc_id);
  USE_SHARED_RO(collected_proc_id) = i;
}

/*
 * update the procedure ID, and adjust the reference count at the same time
 */
inline void update_ref_count(DisjointSetMemberT *p, 
		      DisjointSetMemberT old_proc_id,
		      DisjointSetMemberT new_proc_id,
		      DisjointSetMemberT new_proc_id_type)
{
    if (old_proc_id != new_proc_id) {
      if ((--USE_SHARED_RO(ref_count)[old_proc_id]) == 0) {
	USE_SHARED_RO(ref_count)[USE_SHARED_RO(parent)[old_proc_id]]--;
	collect_proc_id(old_proc_id);
      }
      *p = new_proc_id_type;
      USE_SHARED_RO(ref_count)[new_proc_id]++;
    }
}

#endif
