/*
 * $Id: arch-linux.h,v 1.7 1997/09/10 20:08:59 randall Exp $
 */
/*
 * $Log: arch-linux.h,v $
 * Revision 1.7  1997/09/10 20:08:59  randall
 * Updated comments.
 *
 * Revision 1.6  1997/09/10 19:40:20  randall
 * Fixed up some locking stuff.
 *
 * Revision 1.5  1997/08/26 13:40:51  athena
 * Hacked licking section on linux because cilk2c does
 * not understand the __asm__ syntax of linux include files.
 * Linux uses strings containing \n, which are not ANSI standard.
 * I did not want to introduce (more) non-ANSIety in cilk2c.
 *
 * Revision 1.4  1997/08/25 22:23:58  randall
 * Formatting...
 *
 * Revision 1.3  1997/08/25 21:20:05  randall
 * Added user-level locks.
 *
 * Revision 1.2  1997/03/01 17:42:30  athena
 * Massive cleanup and merging --- since I was still using the
 * old repository :-(
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#define CACHE_LINE 64

/* padding, to be used in structures to avoid false sharing */
#define CILK_CACHE_LINE_PAD  char __dummy[CACHE_LINE]

#define CACHE_LINE_ALIGNED(decl) decl  __attribute__((aligned(CACHE_LINE)))

#if (!CILK_LINUX)
#error CILK_LINUX not defined !
#endif

/***********************************************************\
 * Lock implementation
\***********************************************************/

/* a lock variable */
typedef volatile int Cilk_mutex;

/***********************************************************\
 * Machine size
\***********************************************************/

#include <unistd.h>

static inline int Cilk_partition_size(void)
{
     /* 
      * This is only really right for an unloaded machine.
      */
     return 1;
}

/*
 * Get OS page size.
 */
static inline size_t Cilk_get_page_size(void)
{
     return sysconf(_SC_PAGESIZE);
}

/***********************************************************
 * Interaction with unix
 ***********************************************************/
#define Cilk_yield()     { usleep(10000); }

/***********************************************************\
 * Various types of membars
\***********************************************************/

/*
 * The linux box we have seems to be sequentially consistent,
 * so most of these functions are NOPs.
 */

/*
 *  Ensure that all previous memory operations are completed before
 *  continuing.
 */
static inline void Cilk_fence(void)
{
     asm volatile("":::"memory");
}

/*
 *  Ensure that all previous writes are globally visible before any
 *  future writes become visible.
 */
static inline void Cilk_membar_StoreStore(void)
{
     asm volatile("":::"memory");
}

/*
 *  Ensure that all previous writes are globally visible before any
 *  future reads are performed.
 */
static inline void Cilk_membar_StoreLoad(void)
{
     asm volatile("":::"memory");
}

/***********************************************************\
 * Timing functions
\***********************************************************/

#include <sys/times.h>

/* type to hold a high-resolution timer.  Must be an arithmetic type. */
typedef long Cilk_time;

static inline Cilk_time Cilk_get_time(void)
{
     struct tms buf;
     times(&buf);
     
     return (buf.tms_utime + buf.tms_stime);
}

static inline double Cilk_time_to_sec(Cilk_time t)
{
     return (double) t / 100.0;
}

static inline Cilk_time Cilk_get_wall_time(void)
{
     struct tms buf;
     
     return times(&buf);
}

static inline double Cilk_wall_time_to_sec(Cilk_time t)
{
     return (double) t / 100.0;
}

/***********************************************************\
 * Fast user locks (spin locks)
\***********************************************************/
typedef volatile int Cilk_lockvar;

struct Cilk_linux_xchg_dummy { unsigned long a[100]; };
#define Cilk_xg_hack(x) ((struct Cilk_linux_xchg_dummy *)(x))

static inline unsigned long Cilk_linux_xchg(Cilk_lockvar *ptr, Cilk_lockvar x)
{
     __asm__("xchgl %0,%1"
	     :"=r" (x)
	     :"m" (*Cilk_xg_hack(ptr)), "0" (x)
	     :"memory");
     return x;
}

static inline void Cilk_lock_init(Cilk_lockvar *v)
{
     *v = 0;
     Cilk_membar_StoreStore();
}

static inline void Cilk_lock(Cilk_lockvar *v)
{
     while (Cilk_linux_xchg((int *)v, 1) != 0) {
	  while (*v);  /* spin using only reads - reduces bus traffic */
     }
}

static inline void Cilk_unlock(Cilk_lockvar *v)
{
     Cilk_membar_StoreStore();
     *v = 0;
}
