/*
 * $Id: cilk-lib.h,v 1.1.1.1.2.5 1997/06/16 19:52:12 fengmd Exp $
 */
/*
 * $Log: cilk-lib.h,v $
 * Revision 1.1.1.1.2.5  1997/06/16 19:52:12  fengmd
 * Add pragma.
 *
 * Revision 1.1.1.1.2.4  1997/06/16 16:00:38  fengmd
 * Remerge from the base for Version 5.0.1.
 *
 * Revision 1.4  1997/03/01 17:42:42  athena
 * Massive cleanup and merging --- since I was still using the
 * old repository :-(
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#ifndef _CILK_LIB_H
#define _CILK_LIB_H

#pragma nd -

#include <cilk.h>

/***********************************************************\
 * Exported variables
\***********************************************************/

/* Number of processors Cilk is running on */
DECLARE_SHARED_RO_VARIABLE(extern int, active_size, );
#define Cilk_active_size USE_SHARED_RO(active_size)

/***********************************************************\
 * Memory allocator functions
\***********************************************************/

#include <stdlib.h>
#include <malloc.h>

/*
 * Define C_ functions to be original functions
 */
/* is this portable?  it seems to be... -KHR */
extern void *__builtin_alloca(size_t size);
#define C_alloca __builtin_alloca

static inline void *C_malloc(size_t size)
{
     return malloc(size);
}

static inline void *C_memalign(size_t alignment, size_t size)
{
     return memalign(alignment, size);
}

static inline void C_free(void *s)
{
     free(s);
}

static inline void *C_calloc(size_t nelem, size_t elsize)
{
     return calloc(nelem, elsize);
}

static inline void *C_realloc(void *s, size_t size)
{
     return realloc(s, size);
}

static inline void *C_valloc(size_t size)
{
     return valloc(size);
}

extern void *Cilk_alloca(size_t size);
extern void *Cilk_malloc(size_t size);
extern void *Cilk_malloc_fixed(size_t size);
extern void *Cilk_memalign(size_t alignment, size_t size);
extern void Cilk_free(void *s);
extern void *Cilk_calloc(size_t nelem, size_t elsize);
extern void *Cilk_realloc(void *s, size_t size);
extern void *Cilk_valloc(size_t size);

/* Cilk_exit */
extern void Cilk_exit(int ret);

/***********************************************************\
 * Random number generator
\***********************************************************/

#include <stdlib.h>

/*
 * Define C_ functions to be original functions
 */
static inline int C_rand(void)
{
     return rand();
}
static inline void C_srand(unsigned int seed)
{
     srand(seed);
}

extern int Cilk_rand(void);
extern void Cilk_srand(unsigned int seed);

/***********************************************************\
 * C's original abort
\***********************************************************/

static inline void C_abort(void)
{
     abort();
}

/* Debugging support */
#ifdef	NDEBUG
#define Cilk_assert(ex) ((void)0)
#else
#define Cilk_assert(ex) \
      (void)((ex) || (Cilk_die(USE_SHARED_RO(assertion_failed_msg), \
             #ex, __LINE__, __FILE__), 0))
#endif

#pragma nd +

#endif				/* _CILK_LIB_H */
