/*
 * $Id: invoke-main.c,v 1.1.1.1.2.2 1997/06/16 16:00:48 fengmd Exp $
 */

/*
 * $Log: invoke-main.c,v $
 * Revision 1.1.1.1.2.2  1997/06/16 16:00:48  fengmd
 * Remerge from the base for Version 5.0.1.
 *
 * Revision 1.2  1997/03/01 17:42:54  athena
 * Massive cleanup and merging --- since I was still using the
 * old repository :-(
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#include <cilk.h>
#include <cilk-lib.h>
#include <cilk-internal.h>
#include <cilk-cilk2c.h>
#include <stdio.h>

DECLARE_SHARED_VARIABLE(extern int, done, );

/************************************************************
 *  Initial thread
 ************************************************************/

/* This is a hand-compiled procedure that calls cilk_main */
struct invoke_main_frame {
     StackFrame header;
     int argc;
     char **argv;
     int retval;
     int res; 
} ;

/* you are not supposed to understand this */
extern int cilk_main(int argc, char *argv[]);

struct invoke_main_catch_inlet_args {int res;};

/* 
 * This inlet is created for the purpose of Cilk_exit(). 
 * It sets retval to the exit value, and calls abort.
 * So, when this inlet is excuted, the user's cilk_main
 * should be aborted as a child of invoke main.
 */
static void invoke_main_catch_inlet(struct invoke_main_frame *_frame,
				     struct invoke_main_catch_inlet_args *inletargs,
				     void* inletresult)
{
     USE_SHARED(retval) = inletargs->res;
     _ABORT_STANDALONE();
}



static void invoke_main_slow(struct invoke_main_frame *_frame)
{
     int argc;
     char **argv;
     int tmp;

     _START_THREAD_SLOW();
     switch (_frame->header.entry) {
     case 1: 
	  goto _sync1;
     case 2: 
	  goto _sync2;
     }
     
     
     Cilk_user_work = 0;
     Cilk_user_critical_path = 0;
     WHEN_CILK_TIMING(_frame->header.cp = (Cilk_time) 0);
     WHEN_CILK_TIMING(_frame->header.work = (Cilk_time) 0);
     WHEN_CILK_TIMING(_frame->header.mycp = (Cilk_time) 0);
     WHEN_CILK_TIMING(USE_PRIVATE(last_cp_time) = Cilk_get_time());

     argc = _frame->argc;
     argv = _frame->argv;

     _frame->header.index = (char *) &_frame->retval -
	  (char *) _frame;
     _frame->header.entry=1;
     _BEFORE_SPAWN_SLOW();
     _PUSH_FRAME(_frame);
     tmp = cilk_main(argc,argv);
     _XPOP_FRAME_RESULT(_frame,/* return nothing */,tmp);
     _AFTER_SPAWN_SLOW();

     USE_SHARED(retval) = tmp;

     if (0) {
     _sync1:
	  ;
     }
     _AT_THREAD_BOUNDARY_SLOW();
     
     _BEFORE_SYNC_SLOW();
     _frame->header.entry=2;
     if (_SYNC) {
	  return;
     _sync2:
	  ;
     }
     _AFTER_SYNC_SLOW();
     _AT_THREAD_BOUNDARY_SLOW();

     USE_SHARED(done) = 1;

     WHEN_CILK_TIMING(USE_SHARED(critical_path) = _frame->header.mycp);
     WHEN_CILK_TIMING(USE_SHARED(total_work) = _frame->header.work);

     Cilk_free(_frame);
     return;
}


static ProcInfo invoke_main_sig[]={
     {sizeof(int), sizeof(struct invoke_main_frame ), invoke_main_slow, 0, 0},
     {sizeof(int), 0, invoke_main_catch_inlet,
      sizeof(struct invoke_main_catch_inlet_args ),
      offsetof(struct invoke_main_frame, res)},
     {0, 0, 0, 0, 0}
};


Closure *Cilk_create_initial_thread(int argc, char *argv[])
{
     Closure *t;
     struct invoke_main_frame *f;

     /* create a frame for invoke_cilk_main */
     t = Cilk_Closure_create();
     t->parent = (Closure *) NULL;
     t->status = CLOSURE_READY;

     f = Cilk_malloc(sizeof(struct invoke_main_frame));
     f->header.entry = 0;
     f->header.sig = invoke_main_sig;
     WHEN_CILK_DEBUG(f->header.magic = CILK_STACKFRAME_MAGIC);

     f->argc = argc;
     f->argv = argv;

     t->frame = &f->header;
     return t;
}
